/* $Workfile: AlvisAsmTone.c $       $Revision: 4 $      $Modtime: 6/07/10 8:39a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    AlvisAsmTone.c

    A simple example console application demonstrating the use the following
    devices on DMPs:
    
      OTX_LDEVICE_TONE_DTMF_DETECTOR, 
      OTX_LDEVICE_TONE_DTMF_DIALER, 

      OTX_LDEVICE_TONE_MF_DETECTOR,
      OTX_LDEVICE_TONE_MF_DIALER
    
   The applications works with the following boards combinations:

      Thor-2(4,8)-PCI-Plus / Gimle-16-PCI-Plus + Alvis-ASM (populated with 1,2 or 4 DMPs) 

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 - 16 E1 Lines (depends on type of board) and expects them to be looped 
    with an external loop cable. 

    It also creates various tone device on the DMP, which are used to detect and generate tones.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDmp.h"
#include "OtxEth.h"


#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

#define DEMO_DTMF_DIALLER_TONE_DURATION 8 // 12.75 milliseconds resolution

#define DEMO_TONE_SENDER 1                // if 1: both tone detectors and senders are needed, mostly used on Thor
                                          // if 0: only tone detectors are needed, mostly used on Gimle

#define DEMO_TASK_REF 1L                  // Task Reference for a non-blocking call

#define DEMO_DEFAULT_LOG_FILE "AlvisAsmTone.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_ASM0_HIGHWAY    10
#define DEMO_ASM1_HIGHWAY    11
#define DEMO_ASM2_HIGHWAY    12
#define DEMO_ASM3_HIGHWAY    13

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;        // T1 or E1 mode
  OtxDspSpmE      eProgramModule;   // Standard Program Module to be loaded in the DSP (from the Lib)
  char           *pszDspOutFile;    // Optionally load the DSP program direct from an OUT file (this file name)
  int             boardNo;
  int             coDsp;
  int             coLi;
  OtxMfToneE      eMfToneType;      // Type of MF Tone to detect
  OTX_BOOL        bUseDspDebugger;
  OTX_BOOL        bLiLoop;

  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;

  OTX_INT32       nDtmfDiallerToneDuration;
  char            *pszDongleType;

  int             nLinkTimeout;

} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoOpenAlvisAsmBoard(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static int DemoCheckDspFamily(OTX_UINT32 nVidarType);
static void* DemoIdleFunction( void *lpParameter );

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
FILE       *g_pDspOutFile = NULL;
DemoEnvT    g_demoEnv;
char *g_sqNumber0 = "0123456789ABCD*#";
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs
OTX_UINT32  g_eDevType = OTX_DEVICE_ALVIS_ASM;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[4];  
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[16];
  OTX_HANDLE hDsp[4];
  OTX_HANDLE hTss;
  OTX_HANDLE hEthSw;
} g_hPhysDevs;


struct Pipes {
  OTX_HANDLE hDspLiDtmfDialer[4];
  OTX_HANDLE hLiDspDtmfDetector[4];
  OTX_HANDLE hDspLiMfDialer[4];
  OTX_HANDLE hLiDspMfDetector[4];
  
  OTX_HANDLE hLiIn[16];
  OTX_HANDLE hLiOut[16];
} g_hPipes;


struct LogicDevs {
  OTX_HANDLE hDtmfDetector[4];
  OTX_HANDLE hDtmfDialler[4];
  OTX_HANDLE hMfDialler[4];
  OTX_HANDLE hMfDetector[4];
} g_hLogicDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DTMF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(result);						

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}


/*******************************************************************************

  @func
    Open Alvis-ASM board and run SPM on DMP devices.

  @comm All DMP devices uses DSP handles and work similar as DSP.

*******************************************************************************/
void DemoOpenAlvisAsmBoard(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  coDmp = 4;
  OTX_HANDLE hDspHandle[4];
  OTX_INT32  nHandleIndex = 0;

  for (i = 0; i < 4; i++) {
    hDspHandle[i] = OTX_INVALID_HANDLE_VALUE;
  }

  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_ALVIS_ASM, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(result);

  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_ETHERNET_SWITCH, 0, 0, g_hEventQueue, &g_hPhysDevs.hEthSw); DemoCheckRc(result);
      
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    char szDevString[32];
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DMP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(result);
    DemoLogf("DSP#%d: %s\n", i, szDevString);
  }

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DMP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);
  }

  // enable physical devices
  result = OtxDrvEnable(g_hPhysDevs.hEthSw, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  OsaThreadSleep( g_demoEnv.nLinkTimeout );
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  OsaThreadSleep(1000);

  // ALVIS ASM supports 1 to 4 DMP devices. The following code 
  // will determine which DMP is installed.
  if(g_demoEnv.pszDspOutFile){
    for (i = 0; i < coDmp; i++) {
	    
      result = OtxDmpRunCodecEngineServer(g_hPhysDevs.hDsp[i], g_demoEnv.pszDspOutFile, "otx.ce.spm1");
      if( OTX_FAILURE( result ) ) {
        if( OTX_RESULT_ERRORCODE( result ) != OTX_RES_FAIL ) {
          DemoCheckRc(result);
        }
        // Just to make a demo program generic for 4 DSPs. 
        // The ALVIS_ASM can have 1 - 4 DMP installed.
        // Mark as invalid here and it will be configured later.
        g_hPhysDevs.hDsp[i] = OTX_INVALID_HANDLE_VALUE; 
        hDspHandle[nHandleIndex] = OTX_INVALID_HANDLE_VALUE; 
        DemoLogf("DMP#%d Not found\n", i);
      } else {
        DemoLogf("DMP#%d CODEC Engine Server running index %d\n", i, nHandleIndex);
        // Save off this handle it is valid and will be used 
        // to configure any DMP that is not available.
        hDspHandle[nHandleIndex] = g_hPhysDevs.hDsp[i];
      }
      nHandleIndex++;
    
    }
  
  }else{

    for (i = 0; i < coDmp; i++) {
      result = OtxDmpRunStandardProgramModule(g_hPhysDevs.hDsp[i], OTX_DMP_SPM_OTXSPM2);
      if( OTX_FAILURE( result ) ) {
        if( OTX_RESULT_ERRORCODE( result ) != OTX_RES_FAIL ) {
          DemoCheckRc(result);
        }
        // Just to make a demo program generic for 4 DSPs. 
        // The ALVIS_ASM can have 1 - 4 DMP installed.
        // Mark as invalid here and it will be configured later.
        g_hPhysDevs.hDsp[i] = OTX_INVALID_HANDLE_VALUE; 
        DemoLogf("DMP#%d Not found\n", i);
      } else {
        // Save off this handle it is valid will be used to configure
        // any DMP that is not available.
        hDspHandle[nHandleIndex] = g_hPhysDevs.hDsp[i];
      }
      nHandleIndex++;

    }
 
  }

  nHandleIndex = 0;
  g_demoEnv.coDsp = 0;

  // Look for DMP handle that is invalid and map the devices
  // to a DMP that is installed. ALVIS ASM may have a single 
  // DMP that will support all logical devices.
  for (i = 0; i < coDmp; i++) {
    
    if(g_hPhysDevs.hDsp[i] == OTX_INVALID_HANDLE_VALUE) {
      
      DemoLogf("DMP#%d handle is not available\n", i );
    
      for (; nHandleIndex < coDmp; nHandleIndex++ ) {
      
        //DemoLogf("nHandleIndex = %d \n", nHandleIndex );

        if(hDspHandle[nHandleIndex] == OTX_INVALID_HANDLE_VALUE) {
        
        } else {
          
          DemoLogf("DMP handle #%d is map to DMP#%d\n", i, nHandleIndex);
          g_hPhysDevs.hDsp[i] = hDspHandle[nHandleIndex];

          nHandleIndex++;
          g_demoEnv.coDsp++;
          break;
        }
      
      } // for

    }

  }

}


/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32  nQuadE1T1; // only used for Thor-8-PCI-Plus, Gimle-16-PCI-Plus and Gimle-16-PCIe
  OTX_UINT32 nVidarBoardType = 0;


  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE:
    DemoLogf("Opening Gimle-16-PCIe board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      // On the Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;

  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    {
      // On the Gimle-16-PCI-Plus board the E1T1 devices are hosted by 4 quad E1T1 devices. Open those four devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    {   
      // Get a handle to the ASM Daughter Board
      OTX_UINT32 eDevType;
      OTX_ITERATOR itrChildTypes;
      
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {

        case OTX_DEVICE_ALVIS_ASM:
          DemoLogf("Opening Alvis-ASM board\n");
          g_eDevType = OTX_DEVICE_ALVIS_ASM;
          DemoOpenAlvisAsmBoard();
          return;

        default:
          break;
        }
          
        g_eDevType = nVidarBoardType;

        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(result);
        }
      
      } //while
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        char szDevString[32];
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
        result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hDsp[i], OTX_ATTR_DEVICE_STRING, szDevString, 32); DemoCheckRc(result);
        DemoLogf("DSP#%d: %s\n", i, szDevString);
      } 
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);

  }


  // Load DSP Program
  if (g_demoEnv.pszDspOutFile) {

    DemoCheckDspFamily(nVidarBoardType);

    // Load from file
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.pszDspOutFile,  nDsp);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(result);

    }

  } else {
    
    // Load DSP Program from OtxDrv.lib
    for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
      switch (g_demoEnv.eProgramModule) {
      case OTX_DSP_SPM_OTXSPM2:
        DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", nDsp);
        break;
      default:
        DemoLogf("Unknown DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
        OtxDrvReleaseLib();
        OsaConsoleExit();
        exit(1);
      }
      
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(result);

    }
  }


  if (g_demoEnv.bUseDspDebugger) {
    int c;
    DemoLogf("************************************************************\n\n");
    DemoLogf("Now please load and run DSP program to all of the DSPs\n");
    DemoLogf("in your DSP Debugger.\n\n");
    DemoLogf("After all DSPs are running:\n");
    DemoLogf("  Press 'r' to resynchronize the communication between the DSPs and the host\n");
    DemoLogf("  Press 's' to skip the resynchronization step\n");
    c = OsaConsoleGetCh();
    DemoLogf("\n");

    if (c == 'r' || c == 'R') {
      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }
  }

}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  if(g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) {
    for (i = 0; i < g_demoEnv.coLi; i++) {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1){ // T1
        // Configure T1 Transceivers
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
        // Set them in T1 mode
        result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(result);
      }
      else{  // E1
        // Configure E1 Transceivers
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
        // Set them in E1 mode
        result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(result);
      }

      // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
      result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(result);

      // Set line termination mode
      {

        DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
        switch (g_demoEnv.eTermMode) {
        case OTX_T1E1_TERM_120_OHM:
          DemoLogf("OTX_T1E1_TERM_120_OHM\n");
          break;
        case OTX_T1E1_TERM_100_OHM:
          DemoLogf("OTX_T1E1_TERM_100_OHM\n");
          break;
        case OTX_T1E1_TERM_75_OHM:
          DemoLogf("OTX_T1E1_TERM_75_OHM\n");
          break;
        case OTX_T1E1_TERM_HIGHZ:
          DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
          break;
        default:
          DemoLogf("Invalid Termination mode.\n");
          break;
        }

        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(result);
      }

      // Set receive threshold
      DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(result);
      { // Read back the threshold
        OTX_UINT32 nRxThreshold;
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(result);
        DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
      }
    
      // Set Monitor (-20dB) mode
      DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
      result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(result);
    
      // Turning front end Amplifier ON/OFF      
      DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(result);
    }

  } else {
  
    // Configure Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {

      // Set termination for the T1/E1 Front End
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
      } else {
        // E1 mode
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
      }
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);

      if (g_demoEnv.bLiLoop) {
        // Tell the E1T1 device to loop its internal highway back to the internal highway
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[i], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(result);
      }
      
    }
  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // only used for Thor-8-PCI-Plus, Gimle-16-PCI-Plus and Gimle-16-PCIe

          
  if(g_eDevType != OTX_DEVICE_ALVIS_ASM) {
    result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // If board type is Thor-8-PCI-Plus then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS)) {
      
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }
    

  // If board type is Gimle-16-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) {

    for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }
  

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    DemoLogf("Create logical device on DSP #%d \n", i);

    // Create DTMF Detector Logical Devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hDtmfDetector[i]); DemoCheckRc(result);

    // Create MF Detector Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_MF_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hMfDetector[i]); DemoCheckRc(result);


#if DEMO_TONE_SENDER
    // Create DTMF Dialer Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_DTMF_DIALER, g_hEventQueue, i, &g_hLogicDevs.hDtmfDialler[i]); DemoCheckRc(result);

    // Create MF Dialer Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_MF_DIALER, g_hEventQueue, i, &g_hLogicDevs.hMfDialler[i]); DemoCheckRc(result);
#endif

  }

}


/*******************************************************************************

  @func
    Close Logical Devices

*******************************************************************************/
void DemoCloseLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    // Close DTMF Detector Logical Devices
    result = OtxDrvClose(g_hLogicDevs.hDtmfDetector[i]); DemoCheckRc(result);

    // Close MF Detector Logical Devices
    result = OtxDrvClose(g_hLogicDevs.hMfDetector[i]); DemoCheckRc(result);


#if DEMO_TONE_SENDER
    // Close DTMF Dialers Logical Devices
    result = OtxDrvClose(g_hLogicDevs.hDtmfDialler[i]); DemoCheckRc(result);
      
    // Close MF Dialer Logical Devices
    result = OtxDrvClose(g_hLogicDevs.hMfDialler[i]); DemoCheckRc(result);
#endif

  }

}


/*******************************************************************************

  @func
    Close Physical Devices

*******************************************************************************/
void DemoClosePhysicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // close DSP
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvClose(g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqLiTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];
  OTX_INT32 nTsLi;
  OTX_INT32 nTsDsp;
  OTX_UINT32 coTsPerDspHwy;
  OTX_UINT32 nDspHighwaySpeed;

  // Create pipe for tone devices
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDspDtmfDetector[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDspMfDetector[i]); DemoCheckRc(result);

#if DEMO_TONE_SENDER
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLiDtmfDialer[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLiMfDialer[i]); DemoCheckRc(result);
#endif

  }                  

  // Connect pipes for the DTMF Dialer and Detector

  for (i = 0; i < g_demoEnv.coDsp; i++) {
      
      nLi = i;

      nDsp = nLi;
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
      coTsPerDspHwy = nDspHighwaySpeed / 64;

      DemoLogf("DSP[%d] physical handle = %X \n", nDsp, g_hPhysDevs.hDsp[nDsp]);

      // Connect Pipes
      nTsLi  = 1;   // demo only

      if (g_eDevType == OTX_DEVICE_ALVIS_ASM) {
        // ALVIS_ASM supports 1 to 4 DMP, so a single DMP will need to support
        // all the logical devices. Map each logical device to a different TS.
        nTsDsp = nTsLi;
      }

      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
      sqDspTimeSlotMask[nTsDsp] = 0xFF;

      DemoLogf("Connecting DTMF device[%d]: nLi=%d, nTsLi=%.2d, nDsp=%d, nTsDsp=%.3d\n", nLi, nLi, nTsLi, nDsp, nTsDsp);

      result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);      
      result = OtxDrvConnectPipe(g_hPipes.hLiDspDtmfDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);

#if DEMO_TONE_SENDER
      result = OtxDrvConnectPipe(g_hPipes.hDspLiDtmfDialer[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
      result = OtxDrvConnectPipe(g_hPipes.hDspLiDtmfDialer[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
#endif
  
  }

  // Connect pipes for the MF Dialer and Detector

  for (i = 0; i < g_demoEnv.coDsp; i++) {
      
      nLi = i;

      nDsp = nLi;
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
      coTsPerDspHwy = nDspHighwaySpeed / 64;

      DemoLogf("DSP[%d] physical handle = %X \n", nDsp, g_hPhysDevs.hDsp[nDsp]);

      // Connect Pipes
      nTsLi  = 5;   // demo only

      if (g_eDevType == OTX_DEVICE_ALVIS_ASM) {
        // ALVIS_ASM supports 1 to 4 DMP, so a single DMP will need to support
        // all the logical devices. Map each logical device to a different TS.
        nTsDsp = nTsLi;
      }

      memset(sqLiTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqLiTimeSlotMask[nTsLi] = 0xFF;
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
      sqDspTimeSlotMask[nTsDsp] = 0xFF;

      DemoLogf("Connecting MF device[%d]: nLi=%d, nTsLi=%.2d, nDsp=%d, nTsDsp=%.3d\n", nLi, nLi, nTsLi, nDsp, nTsDsp);

      result = OtxDrvConnectPipe(g_hPipes.hLiDspMfDetector[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);      
      result = OtxDrvConnectPipe(g_hPipes.hLiDspMfDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);

#if DEMO_TONE_SENDER
      result = OtxDrvConnectPipe(g_hPipes.hDspLiMfDialer[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
      result = OtxDrvConnectPipe(g_hPipes.hDspLiMfDialer[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqLiTimeSlotMask, 32); DemoCheckRc(result);
#endif

  }


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------

    // Create the following Connections in receive direction:
    // Li ------> DtmfDetector
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspDtmfDetector[i], -1, g_hLogicDevs.hDtmfDetector[i], -1); DemoCheckRc(result);

    // Li ------> MfDetector
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspMfDetector[i], -1, g_hLogicDevs.hMfDetector[i],   -1); DemoCheckRc(result);

#if DEMO_TONE_SENDER
    // Make the following connections in transmit direction:
    // DtmfDialler   -----> Li
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDtmfDialler[i], -1, g_hPipes.hDspLiDtmfDialer[i], -1); DemoCheckRc(result);
    
    // MfDialler -----> Li
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hMfDialler[i], -1, g_hPipes.hDspLiMfDialer[i], -1); DemoCheckRc(result);
#endif

  }
  
}



/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    // Initialize the DTMF Detectors
    result = OtxToneDtmfDetectorInit(g_hLogicDevs.hDtmfDetector[i], 
                                     g_demoEnv.law, 
                                     OTX_TRUE); DemoCheckRc(result);

    DemoLogf("DTMF Detector[%d]: Tone duration = %d\n", i, g_demoEnv.nDtmfDiallerToneDuration);

    if (i == 0) // read once 
    {
      OTX_UINT32 nThreshold;

      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);

      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);

      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST, &nThreshold); DemoCheckRc(result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_TWIST        = %d\n", i, nThreshold);

      result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hMfDetector[i], OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD, &nThreshold); DemoCheckRc(result);
      DemoLogf("MF Detector[%d]: OTX_ATTR_MF_DETECTOR_THRESHOLD_RELATIVE_3RD = %d\n", i, nThreshold);
    }

    // Initialize the MF Tone Detectors
    result = OtxToneMfDetectorInit(g_hLogicDevs.hMfDetector[i], 
                                   g_demoEnv.law,
                                   g_demoEnv.eMfToneType, // MF Tone type (OTX_TONE_MF_R1, OTX_TONE_MF_R2_FORWARD, OTX_TONE_MF_R2_BACKWARD)
                                   OTX_TRUE,              // If OTX_TRUE, generate events and store the received digits in the Event QueusFifo, If OTX_FALSE Don't store. 
                                   OTX_FALSE,             // If OTX_TRUE the detector generates an event when a tone ends    
                                   25);                   // Minimum duration (in milliseconds) required to validate a digit 
    DemoCheckRc(result);

#if DEMO_TONE_SENDER
    // Initialize the DTMF Diallers
    result = OtxToneDtmfDiallerInit(g_hLogicDevs.hDtmfDialler[i], 
                                    g_demoEnv.law, 
                                    g_demoEnv.nDtmfDiallerToneDuration, 
                                    g_demoEnv.nDtmfDiallerToneDuration, 
                                    OTX_TONE_GAIN_0DB); DemoCheckRc(result);
    
    // Initialize the MF Diallers
    result = OtxToneMfDiallerInit(g_hLogicDevs.hMfDialler[i], 
                                  g_demoEnv.law, 
                                  g_demoEnv.eMfToneType,
                                  50, 
                                  25, 
                                  OTX_TONE_GAIN_0DB); DemoCheckRc(result);

#endif

  }
}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < g_demoEnv.coDsp; i++) {

    result = OtxDrvEnable(g_hLogicDevs.hDtmfDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    result = OtxDrvEnable(g_hLogicDevs.hMfDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

#if DEMO_TONE_SENDER
    result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    result = OtxDrvEnable(g_hLogicDevs.hMfDialler[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
#endif
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        switch (pEventData->m_nParam) {
        case 0:  // 0  (941Hz/1336Hz)
        case 1:  // 1  (697Hz/1209Hz)
        case 2:  // 2  (697Hz/1336Hz)
        case 3:  // 3  (697Hz/1477Hz)
        case 4:  // 4  (770Hz/1209Hz)
        case 5:  // 5  (770Hz/1336Hz)
        case 6:  // 6  (770Hz/1477Hz)
        case 7:  // 7  (852Hz/1209Hz)
        case 8:  // 8  (852Hz/1336Hz)
        case 9:  // 9  (852Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 10: // A  (697Hz/1633Hz)
        case 11: // B  (770Hz/1633Hz)
        case 12: // C  (852Hz/1633Hz)
        case 13: // D  (941Hz/1633Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key %C \n", pEventData->m_nSourceId, pEventData->m_nParam - 10 + 'A');
          break;
        case 14: // *  (941Hz/1209Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '*'\n", pEventData->m_nSourceId);
          break;
        case 15: // #  (941Hz/1477Hz)
          DemoLogf("DtmfDetector[%d] detected DTMF key '#'\n", pEventData->m_nSourceId);
          break;
        default: 
          DemoLogf("DtmfDetector[%d] Unrecognized DTMF tone (%d)\n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("MfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("MfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_MF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_DIGIT) {
        DemoLogf("MfDetector[%d] detected MF digit %.2d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else if (pEventData->m_nCode == OTX_MF_DETECTOR_EC_VERBOSE_STATUS){
        DemoLogf("MfDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        if (pEventData->m_nParam) {
          DemoLogf("MfDetector[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("MfDetector[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
        }
      }
    } else {
      DemoLogf("MfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_ETHERNET_SWITCH:
    {
      char *szStatus;
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      switch (pEventData->m_nCode) {
      case OTX_ETH_EC_LEARNED_MAC_ADDRESS:
        // This event carry additional information so we treat it separately
        //DemoLogf("EthSw[] Event: %s (0x%.8X 0x%.8X)\n", msgStr, pEventData->m_nRequestId, pEventData->m_nParam);
        DemoLogf("EthSw[] Event: %s\n", msgStr);
        switch (pEventData->m_nParam & 0x00007000) {
        case 0x00001000:
          szStatus = "Dynamic MAC Address";
          break;
        case 0x00003000:
          szStatus = "Static MAC Address";
          break;
        case 0x00005000:
          szStatus = "Static MAC Address with source and destination filter";
          break;
        case 0x00006000:
          szStatus = "Static MAC Address with source filter";
          break;
        case 0x00007000:
          szStatus = "Static MAC Address with destination filter";
          break;
        default:
          szStatus = "Invalid or Reserved";
          break;
        }
        DemoLogf("  Status: %s\n", szStatus);
        DemoLogf("  VLANID: %d\n", pEventData->m_nParam & 0xFFF);
        DemoLogf("  MAC   : %.2X-%.2X-%.2X-%.2X-%.2X-%.2X\n", 
          (OTX_UINT8)((pEventData->m_nRequestId >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 16) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 8) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 0) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 16) & 0xFF)
          );
        break;
      case OTX_ETH_EC_DELETED_MAC_ADDRESS:
        // This event carry additional information so we treat it separately
        DemoLogf("EthSw[] Event: %s\n", msgStr);
        DemoLogf("  VLANID: %d\n", pEventData->m_nParam & 0xFFF);
        DemoLogf("  MAC   : %.2X-%.2X-%.2X-%.2X-%.2X-%.2X\n", 
          (OTX_UINT8)((pEventData->m_nRequestId >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 16) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 8) & 0xFF),
          (OTX_UINT8)((pEventData->m_nRequestId >> 0) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 24) & 0xFF),
          (OTX_UINT8)((pEventData->m_nParam >> 16) & 0xFF)
          );
        break;
      case OTX_ETH_EC_QOS_STATUS:
        DemoLogf("EthSw[Port#%d] Event: %s, QoS Status =  0x%.4X\n", pEventData->m_nParam, msgStr, (OTX_UINT16)pEventData->m_nRequestId);
        break;
      default:
        // Other events have the Ethernet Switch port number in pEventData->m_nParam
        DemoLogf("EthSw[Port#%d] Event: %s \n", pEventData->m_nParam, msgStr);
        break;
      }
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->eT1E1Mode        = OTX_T1E1_MODE_T1;
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM2; 
  envP->pszDspOutFile    = NULL;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_PCI;
  envP->coDsp            = 4;
  envP->eMfToneType      = OTX_TONE_MF_R1;
  envP->bUseDspDebugger  = OTX_FALSE;
  envP->bLiLoop          = OTX_FALSE;

  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV

  envP->nDtmfDiallerToneDuration = DEMO_DTMF_DIALLER_TONE_DURATION;
  envP->pszDongleType     = NULL;
  envP->nLinkTimeout      = 3000;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == '9') {   // DSP Out file (overrides the DSP Program module switch)
          envP->pszDspOutFile = argv[i] + 2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == '0') {   // Li Loop
          envP->bLiLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == '2') {          // OtxSpm2.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi  = 2;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi  = 8;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 16;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;
            envP->coLi = 16;
            envP->coDsp = 4;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown Term Mode command option %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // DTMF dialler tone duration
          envP->nDtmfDiallerToneDuration = strtoul(argv[i]+2, NULL, 0);
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}   


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[T|E|4|P|G|9]: Board Type used for the demo, where\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board with Alvis-ASM \n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus with Alvis-ASM Board\n");
  DemoLogf("                  9 = Use Gimle-16-PCIe with Alvis-ASM Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P[2]        : DSP Program module, where\n");
  DemoLogf("                  2 = Use OtxSpm2 DSP module (default)\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced (default)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /X           : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /F<Duration> : DTMF dialler tone duration (Default is %d)\n", DEMO_DTMF_DIALLER_TONE_DURATION);
  DemoLogf(" /0           : Loop Lis (TX->RX)\n");
  DemoLogf(" /9<fileName> : DSP Out File (overrides the /P switch)\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

    DSP program file *.out has a magic number (Target ID) to indicate DSP device:
    
    At Byte 20 (starting from 0), 0x98 indicates C54 device, 
    0x9C indicates C55 device. 

  @comm

*******************************************************************************/
int DemoCheckDspFamily(OTX_UINT32 nVidarType)
{
  int  i, magicNumber;

  magicNumber = 0;

  // open DSP *.out file to read line from
  if ((g_pDspOutFile = fopen(g_demoEnv.pszDspOutFile, "r")) == NULL) {
    DemoLogf("Couldn't open DSP *.out file %s\n", g_demoEnv.pszDspOutFile);
    OsaConsoleExit();
		exit(-1);
  }

  // read in first 21 bytes and place the 21st in magicNumber */   
  for( i = 0; (i < 21 ) && ( feof( g_pDspOutFile ) == 0 ); i++ ) {
    magicNumber = fgetc( g_pDspOutFile );
  }

  DemoLogf("Magic number in %s is 0x%x. \n", g_demoEnv.pszDspOutFile, magicNumber);

  if (magicNumber == 0x98) {    
    if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || 
        (nVidarType == OTX_DEVICE_VIDAR_5x_ASM) || (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_CST) ||
        (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_EX) || (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_PRO) 
      ) {
    } else {
      DemoLogf("%s is not a C55x program. \n", g_demoEnv.pszDspOutFile);
      fclose( g_pDspOutFile );
      OtxDrvReleaseLib();
      OsaConsoleExit();
      exit(1);
    }
  }
 
  if (magicNumber == 0x9C) {    
    if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || 
        (nVidarType == OTX_DEVICE_VIDAR_5x_ASM) || (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_CST) ||
        (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_EX) || (nVidarType == OTX_DEVICE_VIDAR_5X4_ASM_PRO) ) {
      
      DemoLogf("%s is not a C54x program. \n", g_demoEnv.pszDspOutFile);
      fclose( g_pDspOutFile );
      OtxDrvReleaseLib();
      OsaConsoleExit();
      exit(1);
    }
  }

  fclose( g_pDspOutFile );

  return 0;
}

/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_HANDLE hCurrentLi = 0;
  static OTX_INT32  nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis, Max Lines %d, DSP %d\n", g_demoEnv.coLi, g_demoEnv.coDsp);
    DemoLogf("===  ========= Current Li %d\n", nCurrentLi);
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" d   Disable board interrupts\n");
    DemoLogf(" e   Enable board interrupts\n");
    DemoLogf(" f   Sending a string of MF digits on TS5 (from current Li)\n");
    DemoLogf(" g   Sending a string of DTMF digits on TS1 (from current Li)\n");
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" L   Query Line Status of all Lis (extended)\n");
    DemoLogf(" r   Disable/Enable DTMF Dialer devices\n");
    DemoLogf(" R   Reset DTMF Dialer device multiple times\n");
    DemoLogf(" x   Disalbe/Enable DTMF Dialer devices on TS1 \n");
    DemoLogf(" z   Set Line Termination Mode\n");
    DemoLogf(" 0-4 Select active Li, used for subsequent commands\n");
    DemoLogf(" q   Quit\n");
    break;

  case 'd':
    DemoLogf("Disabling board interrupts\n");
    OtxDrvDisable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'e':
    DemoLogf("Enabling board interrupts\n");
    OtxDrvEnable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'f':
    {
      OTX_HANDLE hHandle = g_hLogicDevs.hMfDialler[nCurrentLi];
      char *sqDigitString;
       
      switch (g_demoEnv.eMfToneType) {
        case OTX_TONE_MF_R1:
          sqDigitString = "0123456789abcde";
          break;
        case OTX_TONE_MF_R2_FORWARD:
        case OTX_TONE_MF_R2_BACKWARD:
          sqDigitString = "123456789abcdef";
          break;
        default:
          DemoLogf("Unsupported value for g_demoEnv.eMfToneType (%d) \n", g_demoEnv.eMfToneType);
          sqDigitString = "1";
          break;
      }
                
      DemoLogf("Sending digits on Li#%d TS%d '%s'\n", nCurrentLi, 1, sqDigitString);
        
      // Start Dialing MF digit string, Non-Blocking call, will receive Event when completed
      result = OtxToneMfDial(hHandle, (OTX_UINT8 *)sqDigitString, strlen(sqDigitString), DEMO_TASK_REF, OTX_TIME_INFINITY);
      if (OTX_RESULT_CODE(result) == OTX_MF_DIALER_RES_BUSY) {
        DemoLogf("Dialer busy.\n");
      } else {
        DemoCheckRc(result);
        DemoLogf("Done.\n");
      }
        
    }
    break;

  case 'g':
    {
      OTX_INT32 nDigits = strlen(g_sqNumber0);
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi];

      DemoLogf("5 Sending digits on Li#%d TS%d '%s'\n", nCurrentLi, 1, g_sqNumber0);

      // Start Dialing DTMF number, Non-Blocking call, will receive Event when completed
      result = OtxToneDtmfDial(hHandle, (OTX_UINT8 *)g_sqNumber0, nDigits, DEMO_TASK_REF, OTX_TIME_INFINITY); 
      if (OTX_RESULT_CODE(result) == OTX_DTMF_DIALER_RES_BUSY) {
        DemoLogf("Dialer busy.\n");
      } else {
        DemoCheckRc(result);
        DemoLogf("Done.\n");
      }

    }
    break;
  
  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'L':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi[nLi], &sLayer1Status); DemoCheckRc(result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;

  case 'o':
    {
      OtxT1E1LoopE eLoopType;
      OTX_INT32 nLi;

      DemoLogf("Set Loopback mode for Line Interfaces \n");

      DemoLogf("0: All Line Loops off                     \n");
      DemoLogf("1: Line Loop                              \n");
      DemoLogf("2: Remote Loop with Jitter Attenuation On \n");
      DemoLogf("3: Remote Loop with Jitter Attenuation Off\n");
      DemoLogf("4: Local Loop                             \n");
      
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      case 1:
        eLoopType = OTX_T1E1_LINE_LOOP;
        break;
      case 2:
        eLoopType = OTX_T1E1_REMOTE_ON_LOOP_JITTER_ATTEN;
        break;
      case 3:
        eLoopType = OTX_T1E1_REMOTE_OFF_LOOP_JITTER_ATTEN;
        break;
      case 4:
        eLoopType = OTX_T1E1_LOCAL_LOOP;
        break;
      default:
        DemoLogf("Invalid entry. Turning Loop off (OTX_T1E1_LOOPS_OFF)\n");
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      }

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Setting Loopback mode (%d) on Li #%d \n", (int)eLoopType, nLi);
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], eLoopType); DemoCheckRc(result);
      }

      DemoLogf("Done.\n");
    }
    break;

 case 'r':
    {
      OTX_INT32 nDialler;
      
      DemoLogf("Disable/Enable DTMF dialler[0-3]\n");
      for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
        result = OtxDrvDisable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
      for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
        result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
      DemoLogf("Done.");
    }
    break;

 case 'R':
    {
      OTX_INT32 nDialler;
      static OTX_INT32 nLoopTimes = 1;
      OTX_INT32 i;
      char strBuf[12];
      
      DemoLogf("Reseting DTMF dialler[0-3]\n");
      DemoLogf("How many times? (%d)", nLoopTimes);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nLoopTimes = strtol(strBuf, NULL, 0);
      }

      for (i = 0; i < nLoopTimes; i++) {
        for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
          result = OtxDrvDisable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }
        for (nDialler = 0; nDialler < g_demoEnv.coDsp; nDialler++) {
          result = OtxDrvEnable(g_hLogicDevs.hDtmfDialler[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }
      }
      DemoLogf("Done.");
    }
    break;

  case 'x':
    {
      OTX_RESULT result;
      OtxDeviceStateE  eDevState;
      OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi];
      
      result = OtxDrvGetState(hHandle, &eDevState); DemoCheckRc(result);
      if (eDevState == OTX_DEVICE_WORKING) {
        DemoLogf("Disabling sending device on Li#%d TS%d \n" , nCurrentLi, 1);
        result = OtxDrvDisable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      } else {
        DemoLogf("Enabling sending device on Li#%d TS%d \n" , nCurrentLi, 1);
        result = OtxDrvEnable(hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
      
      DemoLogf("Done.\n");
    }
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(result);
    }
    break;

  case '0':
    // Select T1/E1 line
    nCurrentLi = 0;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '1':
    // Select T1/E1 line
    nCurrentLi = 1;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '2':
    // Select T1/E1 line
    nCurrentLi = 2;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '3':
    // Select T1/E1 line
    nCurrentLi = 3;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 i = 0;

  OsaConsoleInit();

  DemoLogf("AlvisAsmTone, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DTMF tone example program on OTX Alvis-ASM.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("MF Tone Type is");
  switch (g_demoEnv.eMfToneType) {
  case OTX_TONE_MF_R1:
    DemoLogf("R1\n");
    break;
  case OTX_TONE_MF_R2_FORWARD:
    DemoLogf("R2 Forward\n");
    break;
  case OTX_TONE_MF_R2_BACKWARD:
    DemoLogf("R2 Backward\n");
    break;
  default:
    DemoLogf("\"Unknown (%d)\" \n", g_demoEnv.eMfToneType);
    break;
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  if (g_hPhysDevs.hVidarAsmController0) {

    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  }

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();
  DemoLogf("\nDemoConnectDevices\n");

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();
  DemoLogf("\nDemoInitializeLogicalDevices\n");

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();
  DemoLogf("\nDemoEnableLogicalDevices\n");

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);
  DemoLogf("\nOsaThreadStart(DemoIdleFunction)\n");

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Read CPLD Build Numbers from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
      
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  OsaThreadSleep(300); // To let DemoIdleFunction start up
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  // close devices
  DemoCloseLogicalDevices();
  DemoClosePhysicalDevices();

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

